-- Maze.Generator v 0.1
-- (c) Alexandr A Alexeev | http://eax.me/

module Maze.Generator(
  Maze,   -- абстрактный тип данных "лабиринт"
  genMaze -- функция генерации лабиринтов
) where
import Data.List

-- возможные направления движения в заданной клетке лабиринта
data Directions =
  Directions { top :: Bool, right :: Bool, bottom :: Bool, left :: Bool }
  
-- лабиринт
data Maze = Maze [[Directions]]

-- список из трех строк для представления клетки лабиринта
cellView (Directions t r b l) =
  [ " " ++ showBool t ++ " ",
    showBool l ++ "\x2588" ++ showBool r,
    " " ++ showBool b ++ " "
  ]
  where
    showBool True = "\x2588" -- unicode-символ "черный пробел"
    showBool _ = " "

-- экземпляр класса Show для клетки лабиринта
instance Show Directions where
  show d =
    concatMap (++ "\n") $ cellView d
    
-- экземпляр класса Show для лабиринта
instance Show Maze where
  show (Maze m) = concatMap (concatMap (++ "\n")) lines
    where
    lines = map ((map (foldl (++) [])).transpose) viewMatrix
    viewMatrix = map (map cellView) m

-- создаем лабиринт размером w x h, где запрещено любое движение
emptyMaze w h = Maze $ map
  (map (\x -> Directions False False False False))
  [x | x <- [[1..w]], t <- [1..h]]
  
-- получить координаты всех клеток лабиринта размером w x h
genMazeCells w h
  | w <= 0 || h <= 0 = []
  | otherwise = [(x, y) | x <- [0..w-1], y <- [0..h-1]]

-- "прорезаем" в лабиринте путь, заданный по координатам клеток
pavePath (Maze m) (from:to:xs) =
  pavePath (Maze $matrixReplaceAt x1 y1 d1' $matrixReplaceAt x2 y2 d2' m) (to:xs)
  where
    -- координаты текущей и следующей клетки
    (x1, y1, x2, y2) = (fst from, snd from, fst to, snd to)
    -- в каком направлении был сделан шаг?
    stepTop    = (x1 == x2) && (y1 - y2 == 1)
    stepBottom = (x1 == x2) && (y2 - y1 == 1)
    stepLeft   = (y1 == y2) && (x1 - x2 == 1)
    stepRight  = (y1 == y2) && (x2 - x1 == 1)
    
    d1 = m !! y1 !! x1 -- текущая клетка
    d2 = m !! y2 !! x2 -- следующая клетка
    -- меняем возможные направления движения в текущей клетке
    d1' = Directions {
      top = (top d1) || stepTop, right = (right d1) || stepRight,
      bottom = (bottom d1) || stepBottom, left = (left d1) || stepLeft
    }
    -- меняем возможные направления движения в следующей клетке
    d2' = Directions {
      top = (top d2) || stepBottom, right = (right d2) || stepLeft,
      bottom = (bottom d2) || stepTop, left = (left d2) || stepRight
    }
    -- замена idx'ового элемента списка lst на itm
    replaceAt idx lst itm =
      (\(a, _:b) -> a++[itm]++b) $ splitAt idx lst
    -- замена элемента (x, y) матрицы mtrx на itm
    matrixReplaceAt x y itm mtrx =
      replaceAt y mtrx $ replaceAt x (mtrx !! y) itm
pavePath m _ = m

-- генерируем все возможные пути длинной _ровно_ pathLen из заданной клетки
genPathsFrom currCell freeCells pathLen
  | pathLen <= 1 || freeCells == [] = [[currCell]]
  | otherwise = map ((:) currCell) $ foldl (++) [] [
      genPathsFrom nextCell (filter (/= nextCell) freeCells) (pathLen-1)
        | nextCell <- freeCells, areNearby currCell nextCell
    ]
  where
  -- являются ли две клетки соседними?
  areNearby (x1,y1) (x2,y2) = (abs(x1-x2) + abs(y1-y2) == 1)

-- генерируем все возможные пути из startCell длинной _до_ maxLen
genAllPathsFrom startCell freeCells maxLen = foldl (++) [] [
    genPathsFrom startCell (filter (/= startCell) freeCells) pathLen
      | pathLen <- [2..maxLen]
  ]

-- генерируем случайный лабиринт размером w x h для заданного randSeed
genMaze :: Int -> Int -> Int -> Maze
genMaze w h randSeed =
  genMaze' randSeed (emptyMaze w h)(genMazeCells w h \\ [(x,y)]) [(x,y)]
  where
    (x, y) = (w `div` 2, h `div` 2)

genMaze' randSeed currMaze freeCells tailCells
  | freeCells == [] = currMaze
  | otherwise =
      genMaze' nextRandSeed
        -- накладываем маршрут на лабиринт
        (pavePath currMaze currPath)
        -- свободные клетки за минусом клеток текущего маршрута
        nextFreeCells
        -- получаем клетки, из которых можно построить маршрут на следующем шаге
        ( filterDeadends $ tailCells ++ (currPath \\ [startCell]) )
  where
    -- отсекаем "тупиковые" ячейки - те, рядом с которыми нет ни одной свободной
    filterDeadends =
      filter (\x -> ((getNearby x) `intersect` nextFreeCells) /= [])
      where
        getNearby (x,y) = [(x-1, y), (x+1,y), (x,y-1), (x,y+1)]
    -- свободные ячейки на следующем шаге рекурсии
    nextFreeCells = freeCells \\ currPath
    -- cлучайным образом выбираем один из маршрутов
    currPath = nextPaths' !! ((abs randSeed) `mod` length nextPaths')
    -- в целях оптимизации рассматриваем только три первых варианта
    nextPaths' = take 3 $ nextPaths
    -- возможные новые маршруты длиной до трех клеток
    nextPaths = genAllPathsFrom startCell freeCells 3
    -- новый маршрут строим из случайно выбранной "хвостовой" клетки
    startCell = tailCells !! ((abs nextRandSeed) `mod` length tailCells)
    -- генерируем следующее "случайное" число
    nextRandSeed = randSeed*1664525 + 1013904223
